from __future__ import annotations
from abc import ABC, abstractmethod
from aiogram import Bot
from meowcorp.status.status import STATUS
from aiogram.types import Message
from models import User, Session
from os import getenv


class Branch(ABC):
    """
    :param status: attr from STATUS_BRANCH
    """
    status: STATUS

    def __init__(self, user: User, db: Session, message: Message):
        """
        Interface for worker of Branch
        All work with telegram goes throw self.bot

        :param user: Current user
        :param db: Current DB session for user
        :param message: Current message
        """
        self.bot: Bot = Bot(token=getenv('BOT_TOKEN'))
        self.user: User = user
        self.db: Session = db
        self.message: Message = message

    @classmethod
    def from_branch(cls, branch: Branch) -> Branch:
        return cls(
            user=branch.user,
            db=branch.db,
            message=branch.message
        )

    @abstractmethod
    async def work(self, *args, **kwargs):
        """
        Handler of current branch

        :return: None
        """
        pass



import sys
import os
import shutil

BANNER = """ Available subcommands:

startproject -> create project
"""


def move_file(directory: str):
    if os.path.isdir(directory):
        files = os.listdir(directory)
        for file in files:
            move_file(os.path.join(directory, file))

    if not directory.endswith('-tpl'):
        return

    path = directory.split(os.sep)
    template_index = path.index("project_tempalte")
    path = path[template_index + 1:]

    new_path = os.getcwd()
    for directory_name in path[:-1]:
        new_path = os.path.join(new_path, directory_name)
        if not os.path.exists(new_path):
            print(f"creating {new_path}")
            os.mkdir(new_path)

    new_path = os.path.join(new_path, path[-1][:-4])
    shutil.copy(directory, new_path)
    print(f"creating {new_path}")


def execute_from_command_line():
    # First version of command execution
    available_commands = ["startproject"]
    args = sys.argv[1:]
    if len(args) == 0 or args[0] not in available_commands:
        return print(BANNER)

    if args[0] == "startproject":
        os.path.normpath(__file__)
        template_dir = os.path.join(
            os.path.dirname(__file__),
            "..",
            "conf",
            "project_tempalte"
        )
        move_file(template_dir)



from aiogram.types import ReplyKeyboardMarkup, KeyboardButton
from aiogram.types import InlineKeyboardMarkup, InlineKeyboardButton
from meowcorp.exceptions.keyboard import *


def create_keyboard(buttons: list, rows: list = None) -> ReplyKeyboardMarkup:
    if rows is None:
        rows = [1] * len(buttons)

    if len(buttons) != sum(rows):
        raise BadButtonsAmount(f"Bad amount of button. Button: {len(buttons)}, rows: {sum(rows)}")

    keyboard = ReplyKeyboardMarkup(resize_keyboard=True)
    new_buttons = [KeyboardButton(x) for x in buttons]

    index = 0
    for amount in rows:
        keyboard.row(*new_buttons[index: index + amount])
        index += amount

    return keyboard


def create_inline_keyboard(buttons: list, rows: list = None) -> InlineKeyboardMarkup:
    if rows is None:
        rows = [1] * len(buttons)

    if len(buttons) != sum(rows):
        raise BadButtonsAmount(f"Bad amount of button. Button: {len(buttons)}, rows: {sum(rows)}")

    keyboard = InlineKeyboardMarkup()
    new_buttons = []
    for x in buttons:
        if isinstance(x, str) or isinstance(x, int):
            new_buttons.append(InlineKeyboardButton(str(x), callback_data=str(x)))
        elif isinstance(x, list) and len(x) == 1:
            new_buttons.append(InlineKeyboardButton(x[0], callback_data=' '))
        else:
            new_buttons.append(InlineKeyboardButton(x[0], callback_data=x[1]))

    index = 0
    for amount in rows:
        if not amount:
            continue
        keyboard.row(*new_buttons[index: index + amount])
        index += amount

    return keyboard



from settings.settings import SEPARATOR
from meowcorp.exceptions.status import SimilarStatuses


class auto:
    pass


class STATUS:
    def __init__(self, raw: str):
        self.raw = raw

    def __call__(self, *args, **kwargs) -> str:
        value = self.raw

        args = list(args) + list(kwargs.values())

        args = list(map(lambda arg: str(int(arg)) if isinstance(arg, bool) else str(arg), args))

        return SEPARATOR.join([value] + args)

    def match(self, user=None, call=None) -> bool:
        if user is not None:
            return user.status.split(SEPARATOR)[0] == self.raw

        if call is not None:
            return call.data.split(SEPARATOR)[0] == self.raw

        return Fasle

    def get_status(self) -> str:
        return self.raw

    def __str__(self) -> str:
        return f"<Status raw={self.raw}>"


class StatusEnum(type):

    def __init__(cls, name, base, attrs):
        super().__init__(name, base, attrs)

        statuses = [name for name, func in attrs.items() if isinstance(func, auto) or func == auto]
        updated = []

        for status in statuses:
            if status.upper() in updated:
                raise SimilarStatuses(f"Got similar statuses: {status}")

            status_class = STATUS(raw=status.upper())
            setattr(cls, status, status_class)

            updated.append(status.upper())



import inspect
from models.models import Base
from settings.settings import SEPARATOR
from meowcorp.exceptions.validation import *


class ArgsValidation:

    def __init__(self):
        self.possible_classes = []
        self._init_classes()

    def _init_classes(self):
        possible_models = __import__('models.models', fromlist=('*',))
        self.possible_classes = []

        for class_object in dir(possible_models):
            attr = getattr(possible_models, class_object)
            if not inspect.isclass(attr):
                continue

            if not issubclass(attr, Base):
                continue

            if attr == Base:
                continue

            self.possible_classes.append(attr)

    def prepare_args(self, class_object, db, user=None, call=None):
        if user is not None:
            args = user.status.split(SEPARATOR)[1:]
        elif call is not None:
            args = call.data.split(SEPARATOR)[1:]
        else:
            return []

        func_args_dict = inspect.getfullargspec(class_object.work)

        func_args = func_args_dict.args[1:]
        defaults = func_args_dict.defaults if func_args_dict.defaults is not None else tuple()

        if len(func_args) > len(args) + len(defaults):
            missing_args = func_args[len(args):]
            if defaults:
                missing_args = func_args[len(args): -len(defaults)]

            missing_args = list(map(lambda x: f"'{x}'", missing_args))

            raise MissingArguments(f'{class_object.__class__.__name__}->work()'
                                   f' missing {len(missing_args)} required args: {", ".join(missing_args)}')

        if len(func_args) < len(args):
            raise ExtraArguments(f'{class_object.__class__.__name__}->work()'
                                 f'takes {len(func_args)} args but {len(args)} was given')

        annotations = func_args_dict.annotations

        for index, arg_name in enumerate(func_args[:len(args)]):
            if arg_name not in annotations:
                continue

            arg_type = annotations[arg_name]

            if arg_type in self.possible_classes:
                instance = db.query(arg_type).get(args[index])
                if instance is None:
                    raise NotFound(f"Instance {arg_type.__name__}"
                                   f" was not found by 'pk = {args[index]}'")

                args[index] = instance
                continue

            if arg_type == int:
                if args[index].isnumeric():
                    args[index] = int(args[index])
                    continue

                raise ConvertException(f"Could not convert {args[index]} to int")

            if arg_type == float:
                if args[index].count('.') <= 1 and \
                        args[index].replace('.', '').isnumeric():
                    args[index] = float(args[index])
                    continue
                raise ConvertException(f"Could not convert {args[index]} to float")

            if arg_type == bool:
                if args[index].isnumeric():
                    args[index] = bool(int(args[index]))
                else:
                    args[index] = bool(args[index])

        return args


args_validation = ArgsValidation()



from telegram.branches import branches
from telegram.callbacks import callbacks

from meowcorp.handlers import args_validation


class Factory:
    def __init__(self, db):
        self.db = db

    async def process_branch(self, user, message):
        for branch in branches:
            if branch.status.match(user=user):
                worker = branch(user, self.db, message)
                args = args_validation.prepare_args(worker, self.db, user=user)
                await worker.work(*args)
                return

    async def process_callback(self, user, call):
        for callback in callbacks:
            if callback.status.match(call=call):
                worker = callback(user, self.db, call)
                args = args_validation.prepare_args(worker, self.db, call=call)
                await worker.work(*args)
                return



class MissingArguments(Exception):
    """
    If worker takes more args than he got
    """
    pass


class ExtraArguments(Exception):
    """
    If worker takes less args than he got
    """
    pass


class NotFound(Exception):
    """
    instance was not found in database by primary key
    """
    pass


class ConvertException(Exception):
    """
    can't convert data to type
    """
    pass



from setuptools import setup, find_packages


with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = [
    "python-dotenv~=0.21.0",
    "sqlalchemy~=1.4.39",
    "psycopg2~=2.9.5",
    "aiogram~=2.23.1"
]

setup(
    name="meowcorp",
    version="0.1.3",
    packages=find_packages(include=['meowcorp', 'meowcorp.*', 'meowcorp.conf.project_template.*']),
    description="Add-on to the library of aiogram",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/EluciferE/meowcorp/",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'meowcorp-admin = meowcorp.commands:execute_from_command_line',
        ]
    }
)
